#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>


static char *strip_end(char *str)
{
	int len = strlen(str);
	for (str += len - 1
		; len > 0 && (isspace(*str) || *str == 26)
		; len--, str--) *str = '\0';
	return str+1;
}


int pam_check_pair(
	const char *user,
	const char *pass);

static int livre (
	const char *user,
	const char *ecole,
	const char *prof,
	const char *_subdir,
	const char *docum,
	const char *original)	// Nom original
{
	int ret = -1;
	struct passwd *p;
	bool found = false;
	while (isspace(*prof)) prof++;
	char subdir[strlen(_subdir)+1];
	strcpy (subdir,_subdir);
	strip_end (subdir);
	setpwent();
	while ((p=getpwent())!=NULL){
		if (strcasecmp(prof,p->pw_gecos)==0){
			// On a trouv le prof
			found = true;
			char path[PATH_MAX];
			if (subdir[0] != '\0' && strcmp(subdir,"/")!=0){
				snprintf (path,sizeof(path)-1,"/ecoles/%s/tp/%s/%s"
					,ecole,prof,subdir);
			}else{
				snprintf (path,sizeof(path)-1,"/ecoles/%s/tp/%s"
					,ecole,prof);
			}
			struct stat st;
			if (stat(path,&st)!=-1){
				char pathf[PATH_MAX];
				time_t ti = time(NULL);
				struct tm *tt = localtime(&ti);
				snprintf (pathf,sizeof(pathf)-1,"%s/%s-%04d-%02d-%02d_%02d-%02d-%02d__%s"
					,path,user,tt->tm_year+1900,tt->tm_mon+1,tt->tm_mday
					,tt->tm_hour,tt->tm_min,tt->tm_sec,original);
				FILE *fin = fopen (docum,"r");
				if (fin != NULL){
					FILE *fout = fopen (pathf,"w");
					if (fout != NULL){
						char buf[10000];
						int n;
						while ((n=fread(buf,1,10000,fin))>0){
							fwrite (buf,1,n,fout);
						}
						ret = fclose (fout);
						if (ret == 0){
							chown (pathf,p->pw_uid,p->pw_gid);
							chmod (pathf,0640);
						}
					}
					fclose (fin);
				}
			}else{
				printf ("Pas de rpertoire de livraison pour le professeur %s<br>\n"
					":%s:<br>\n"
					":%s:<br>\n"
					":%s:<br>\n"
					,prof,path,docum,original);
			}
		}
	}
	if (!found){
		printf ("Aucun compte usager correspondant au professeur :%s:\n",prof);
	}
	return ret;
}
/*
	Vrifie si un rpertoire est valide (ne contient pas de ..)
*/
static bool valid_subpath (const char *subdir)
{
	bool ret = true;
	if (strstr(subdir,"..")!=NULL){
		syslog (LOG_ERR,"Requete %s rejet",subdir);
		ret = false;
	}
	return ret;
}

const int MAXIMUM_ENTRIES=10000;


static int liste_recur (
	const char *path,
	const char *subdir,
	bool recur,
	bool showdir,
	char *tb[],
	int nb,
	int level)
{
	int ret = -1;
	char spath[PATH_MAX];
	if (!valid_subpath (subdir)) return -1;
	if (subdir[0] != '\0'){
		snprintf (spath,sizeof(spath)-1,"%s/%s",path,subdir);
	}else{
		snprintf (spath,sizeof(spath)-1,"%s",path);
	}
	DIR *d = opendir (spath);
	if (d != NULL){
		struct dirent *ent;
		ret = nb;
		while (nb < MAXIMUM_ENTRIES && (ent=readdir(d))!=NULL){
			const char *name = ent->d_name;
			if (name[0] != '.'
				&& (strcmp(name,"Desktop")!=0 || level >0)
				&& (strcmp(name,"Netscape")!=0 || level >0)){
				char sspath[PATH_MAX];
				snprintf (sspath,sizeof(sspath)-1,"%s/%s/%s",path,subdir,name);
				struct stat st;
				if (lstat(sspath,&st)!=-1){
					char ssub[PATH_MAX];
					if (subdir[0] != '\0'){
						snprintf (ssub,sizeof(ssub)-1,"%s/%s",subdir,name);
					}else{
						snprintf (ssub,sizeof(ssub)-1,"%s",name);
					}
					bool is_dir = S_ISDIR(st.st_mode) != 0;
					if (is_dir){
						if (showdir){
							tb[nb++] = strdup(ssub);
						}
						if (recur){
							nb = liste_recur(path,ssub,recur,showdir,tb,nb,level+1);
						}
					}else if (!showdir && S_ISREG(st.st_mode)){
						tb[nb++] = strdup(ssub);
					}
				}
			}
		}
		closedir(d);
		ret = nb;
	}
	return ret;
}

static int cmp_func (const void *p1, const void *p2)
{
	char *s1 = *(char**)p1;
	char *s2 = *(char**)p2;
	return strcasecmp(s1,s2);
}


static int liste (
	const char *path,
	const char *subdir,
	bool recur,
	bool showdir)
{
	if (!valid_subpath (subdir)) return -1;
	char *tb[MAXIMUM_ENTRIES];
	int ret = liste_recur (path,subdir,recur,showdir,tb,0,0);
	if (ret > 0){
		qsort (tb,ret,sizeof(char*),cmp_func);
		for (int i=0; i<ret; i++){
			char *pt = tb[i];
			printf ("%s\n",pt);
			free (pt);
		}
	}
	return ret;
}


struct CTX{
	struct passwd *p;
	struct group *g;
	char ecole[100];
};

/*
	Valide l'usager et le mot de passe
	Met a jour ctx avec l'information sur l'usager
*/
static int livre_getpwnam (
	const char *user,
	const char *pass,
	CTX &ctx)
{
	int ret = -1;
	ctx.p = NULL;
	ctx.g = NULL;
	ctx.ecole[0] = '\0';
	if (pam_check_pair (user,pass)){
		ctx.p = getpwnam (user);
		if (ctx.p != NULL){
			gid_t gid = ctx.p->pw_gid;
			ctx.g = getgrgid (gid);
			if (ctx.g != NULL){
				if (strcmp(user,"root")==0
					|| strcmp(user,"maitre")==0){
					strcpy (ctx.ecole,"0");
					ret = 0;
				}else{
					const char *group = ctx.g->gr_name;
					while (isalpha(*group)) group++;
					if (isdigit(*group)){
						snprintf (ctx.ecole,sizeof(ctx.ecole)-1,"%s",group);
						ret = 0;
					}
				}
			}
		}
	}
	return ret;
}

static bool is_admin_print (CTX &ctx)
{
	bool ret = false;
	FILE *fin = fopen ("/etc/print/admin.conf","r");
	if (fin != NULL){
		char buf[100];
		while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
			int len = strlen(buf);
			while (len > 0 && buf[len-1] <= ' '){
				buf[len-1] = '\0';
				len--;
			}
			if (len > 0){
				if (buf[0] == '@'){
					if (strcmp(buf+1,ctx.g->gr_name)==0){
						ret = true;
						break;
					}
				}else if (strcmp(buf,ctx.p->pw_name)==0){
					ret = true;
					break;
				}
			}
		}
		fclose (fin);
	}
	return ret;
}

/*
	Retourne vrai si le fichier ne contient pas de caractre spciaux
	pouvant nuire  la scurit.
*/
static bool livre_checkfname (const char *original)
{
	bool ret = true;
	if (strchr(original,'/')!=NULL
		|| strchr(original,'(')!=NULL
		|| strchr(original,')')!=NULL
		|| strchr(original,'"')!=NULL
		|| strchr(original,'\'')!=NULL
		|| strchr(original,'`')!=NULL
		|| strchr(original,'*')!=NULL
		|| strchr(original,'?')!=NULL){
		ret = false;
	}
	return ret;
}


int main (int argc, char *argv[])
{
	int ret = -1;
	// Ce programme agit en fonction du nombre de paramtre
	// 2 paramtres: Valide le mot de passe
	// 3 paramtres: Valide le mot de passe et extrait les sous-rpertoires
	//              de livraison pour un prof. Cela veut dire que
	//              /ecoles/num-ecole/tp peut avoir des accs trs restreint
	//				Par exemple, chaque rpertoire de professeur pourrait avoir
	//				des permissions 700. Les tudiant peuvent livrer, mais
	//				ne peuvent pas lire les documents, ni meme savoir quels
	//				existe.
	// 6 paramtres: Valide le mot de passe et fait une livraison
	openlog ("livre_check_pass",LOG_PID,LOG_DAEMON);
	if (argc >= 3){
		const char *user = argv[1];
		CTX ctx;
		if (livre_getpwnam (user,argv[2],ctx)!=-1){
			if (argc == 3){
				printf ("%s\n%s\n",ctx.ecole,ctx.g->gr_name);
				ret = 0;
			}else if (argc == 4){
				if (strcmp(argv[3],"--profliste")==0){
					// On prsente la liste des profs
					// soit la liste de sous-rpertoire dans /ecoles/XX/tp
					char path[PATH_MAX];
					snprintf (path,sizeof(path)-1
						,"/ecoles/%s/tp/",ctx.ecole);
					liste (path,"",false,true);
					ret = 0;
				}else if (strcmp(argv[3],"--adminprint")==0){
					// Confirme si cet usager est un administrateur d'imprimante
					if (is_admin_print (ctx)){
						printf ("ok\n");
						ret = 0;
					}
				}else if (strcmp(argv[3],"--userdirs")==0){
					// Affiche la liste des rpertoires et sous-rpertoires
					// dans le HOME de l'usager
					liste (ctx.p->pw_dir,"",true,true);
				}else if (strcmp(argv[3],"--proftp")==0){
					// Affiche la liste des sous-rpertoire dans le
					// rpertoire tp d'un prof.
					char path[PATH_MAX];
					snprintf (path,sizeof(path)-1
						,"/ecoles/%s/tp/%s",ctx.ecole,ctx.p->pw_gecos);
					liste (path,"",true,true);
					ret = 0;
				}
			}else if (argc == 5){
				if (strcmp(argv[3],"--profdirs")==0){
					// Affiche la liste des sous-rpertoire dans le
					// rpertoire tp d'un prof.
					char path[PATH_MAX];
					snprintf (path,sizeof(path)-1
						,"/ecoles/%s/tp/%s",ctx.ecole,argv[4]);
					liste (path,"",true,true);
					ret = 0;
				}else if (strcmp(argv[3],"--removejob")==0){
					// Efface une tche d'impression
					if (is_admin_print(ctx)){
						char path[PATH_MAX];
						snprintf (path,sizeof(path)-1
							,"/var/print/%s",argv[4]);
						unlink (path);
						ret = 0;
					}
          		}else if (strcmp(argv[3],"--userfiles")==0){
					// Affiche les fichiers d'un rpertoire d'un usager
					ret = liste (ctx.p->pw_dir,argv[4],false,false);
				}else if (strcmp(argv[3],"--userfile")==0){
					// Rcupre un fichier d'un usager
					if (valid_subpath (argv[4])){
						char spath[PATH_MAX];
						snprintf (spath,sizeof(spath)-1,"%s/%s"
							,ctx.p->pw_dir
							,argv[4]);
						struct stat st;
						if (stat(spath,&st)!=-1 && S_ISREG(st.st_mode)){
							FILE *fin = fopen (spath,"r");
							if (fin != NULL){
								char buf[1024*8];
								int n;
								while ((n=fread(buf,1,sizeof(buf),fin))>0){
									fwrite (buf,1,n,stdout);
								}
								fclose (fin);
								ret = 0;
							}
						}
					}
				}
			}else if (argc == 6){
				if (strcmp(argv[3],"--userupload")==0){
					// Recoit un document et le met dans le home de l'usager
					// dans le sous-rpertoire recu.
					const char *original = argv[5];
					if (!livre_checkfname (original)){
						ret = -1;
					}else{	
						char spath[PATH_MAX],pathf[PATH_MAX];
						snprintf (spath,sizeof(spath)-1,"%s/recu",ctx.p->pw_dir);
						mkdir (spath,0770);
						chown (spath,ctx.p->pw_uid,(gid_t)-1);
						// Encode un nom avec la date
						time_t ti = time(NULL);
						struct tm *tt = localtime(&ti);
						snprintf (pathf,sizeof(pathf)-1,"%s/%04d-%02d-%02d_%02d-%02d-%02d__%s"
							,spath,tt->tm_year+1900,tt->tm_mon+1,tt->tm_mday
							,tt->tm_hour,tt->tm_min,tt->tm_sec,original);
						FILE *fin = fopen (argv[4],"r");
						if (fin != NULL){
							FILE *fout = fopen (pathf,"w");
							if (fout != NULL){
								char buf[8*1024];
								int n;
								while ((n=fread(buf,1,8*1024,fin))>0){
									fwrite (buf,1,n,fout);
								}
								ret = fclose (fout);
								if (ret == 0){
									chown (pathf,ctx.p->pw_uid,(gid_t)-1);
								}else{
									syslog(LOG_ERR,"--userupload: Ne peut sauver %s (%m)",pathf);
								}
							}else{
								syslog(LOG_ERR,"--userupload: Ne peut crire %s (%m)",pathf);
							}
							fclose (fin);
						}else{
							syslog(LOG_ERR,"--userupload: Ne peut lire %s (%m)",argv[4]);
						}
					}
				}
			}else if (argc == 7){
				const char *original = argv[6];
				if (!livre_checkfname(original)){
					ret = -1;
				}else{
					ret = livre (ctx.p->pw_gecos,ctx.ecole
						,argv[3],argv[4],argv[5]
						,original);
				}
			}
		}
	}
	return ret;
}

